import argparse
import os
import sys
import dotenv
from .sis import mcp

def main():
    # Load environment variables from .env file first.
    # Command-line arguments will take precedence over these.
    dotenv.load_dotenv()

    parser = argparse.ArgumentParser(description="Start MCP SIS server")
    parser.add_argument("--host", type=str, default="0.0.0.0", help="Host to bind (default: 0.0.0.0)")
    parser.add_argument("--port", type=int, default=3000, help="Port to listen on (default: 3000)")
    parser.add_argument("--transport", type=str, default="stdio", help="Transport type (default: stdio)")
    
    # Set default values from environment variables. CLI arguments will override them.
    parser.add_argument("--username", type=str, default=os.getenv("SIS_USERNAME"), 
                        help="Username for SIS. Overrides SIS_USERNAME in .env file.")
    parser.add_argument("--password", type=str, default=os.getenv("SIS_PASSWORD"), 
                        help="Password for SIS. Overrides SIS_PASSWORD in .env file.")
    
    args = parser.parse_args()

    # After parsing, args will contain credentials from CLI or .env.
    # Now, validate that credentials are set.
    if not args.username or not args.password:
        print("Error: SIS_USERNAME and SIS_PASSWORD must be provided.", file=sys.stderr)
        print("You can set them in a .env file or provide them as command-line arguments (--username, --password).", file=sys.stderr)
        sys.exit(1)

    # Set the final credentials back into the environment. 
    # This ensures the rest of the application (e.g., sis.py) can access them,
    # especially if they were provided via command-line arguments.
    os.environ['SIS_USERNAME'] = args.username
    os.environ['SIS_PASSWORD'] = args.password

    print("Starting SIS-MCP server...")
    mcp.run(transport=args.transport, host=args.host, port=args.port)

if __name__ == "__main__":
    main()